// Copyright (c) 2020 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "source/fuzz/transformation_context.h"

namespace spvtools {
namespace fuzz {

TransformationContext::TransformationContext(
    FactManager* transformation_context,
    spv_validator_options validator_options)
    : fact_manager_(transformation_context),
      validator_options_(validator_options) {}

TransformationContext::~TransformationContext() = default;

}  // namespace fuzz
}  // namespace spvtools
